!+
! Module: epexposure_mod
!
! Authors:  Michael Freyberg - XMM-SSC-MPE
!
! Description: spatial exposure corrections for
!              for EPIC pn scientific observations
!
! Comments:
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!-
module epexposure_mod

  use eputils_mod
  use dal
  use cal
  use oal
  use param
  use errorHandling

  implicit none

  private

! Global variables

  public epexposure

contains

  !---------------------------------------------------------------------
  subroutine epexposure
  !---------------------------------------------------------------------

  character(len=10), parameter :: CodeVersion = '0.14'

  type(DataSetT)                :: infile
  type(TableT)                  :: expo, dlimap, events  ! , inccd
  type(ColumnT)                 :: dlicol(0:1), dliexpco ! , expo_column
  type(ColumnT)                 :: ndsclin1, ndsclin2, ndsclin3, ndsclin4

  real(kind=double), dimension(:),     pointer :: time_exp, time_raw, &
                                                  time_rnd
  integer(kind=Int8),  dimension(:),   pointer :: ccdnr
  integer(kind=Int32), dimension(:),   pointer :: outdlicol
  real(kind=single), dimension(:),     pointer :: dliexpcoCol
  integer(kind=Int32), dimension(:,:,:)        :: dliarray(1:2,1:EPN_MAX_Y,0:12)

  integer(kind=Int16), dimension(:), pointer :: ndsclin1Col, ndsclin2Col, &
                                                ndsclin3Col, ndsclin4Col

  real(kind=single),   dimension(:),  pointer     :: randomnumber => null()
  integer(kind=Int16), dimension(:),  allocatable :: dummy
  integer(kind=Int16), dimension(:,:),allocatable :: ndsclin

  integer(kind=Int16) :: quadrantId, &    ! quadrant identifier [0-3]
                         ccdId, &         ! CCD identifier      [0-2]
                         ccdno            ! overall CCD number  [1-12]
  integer, parameter  :: nccd = 12, nquad = 4

  integer             :: i,j, irow, ir, nrows, mrows, npar, dliodf_tot_quad(0:3), &
                         ic, ic_b(1:nccd), ic_e(1:nccd), &
                         trebin, ccdmode, col_per_quad, col_to_scale, &
                         dliodf_total, dlisas_total, dli_total, ccxdslin(0:nccd), &
                         ndsclinc, an_maxfrc, an_eamip, &
                         ndsclinq, ndsclinq_quad(0:nquad-1), diff, an_chop, &
                         revolution, ic0, ic1, ic2, expo_dlimap(1:EPN_MAX_X), &
                         k, n, ii, noverflow(0:nquad-1), moverflow(0:nquad-1), &
			 nval, ierr, mccd = 0
  real(kind=double)   :: an_dslin, dslin_pmh(0:nquad-1), ndsclina, average, &
                         dlimean2(0:nccd) = 0, dlimedian2(0:nccd) = 0,&
                         dliadev(0:nccd) = 0, dlisdev(0:nccd) = 0, &
                         dliskew(0:nccd) = 0, dlicurt(0:nccd) = 0, &
                         dlimin(0:nccd) = 0, dlimax(0:nccd) = 0, &
                         xval(1:nccd), medmean3, medmedian3, rmin, rmax, &
                         avemean2, avemedian2, adev,sdev,var,skew,curt
  real(kind=double)   :: timedelta, frameTime, T_Counting_Mode
  real(kind=double)   :: t_raw, t_exp, t_rnd, ontime, nframe, fifo_loss, rr

  character(len=1024),dimension(:) :: eventsets(0:2) ! , &
!                                     framesets(0:2), colnames(0:13), &
!                                     coltypes(0:13), attribute(0:11)
! character(len=1024) :: imagingset, mergedname, &
!                        auxfileset, phkfileset, auxfilename, phkfilename, &
!                        expression, filteredname
  character(len=8)    :: extname(0:nccd), colname(0:4), ext_events, contime(0:nccd)
  character(len=4)    :: ExpIDstr
  character(len=10)   :: ObsID
  character(len=12)   :: An_SENDMODE
  character(len=13)   :: OFSTCOR1(0:nccd), ofstref1
  character(len=30)   :: subMode

  integer             :: numberOfCCDs = 0

  logical(kind=bool)  :: hasExtension, isIntermediate, lfirst = .true.
  logical(kind=bool)  :: emptyEvents = .false.
  logical             :: OFSTCOR4(0:nccd), ofstref4
! logical             :: hasOFSTCOR1, hasOFSTCOR4
  logical             :: warningOFSTCOR = .false.
  logical             :: hasExposure(0:nccd) = .false., checkOFSTCOR = .false.
  logical             :: single_ccd_mode, done, lccd(0:nccd), ldisc, &
                         dlimap_overflow = .FALSE., fastmode = .FALSE.

  logical             :: trandom   ! trigger for randomization of time within frame
  logical             :: escreen   ! trigger for removal of extra EXPOSU columns
  logical             :: espatial  ! trigger for determination of spatial exposure inhomogeneities (CCD columns)
  logical             :: lrebin    ! trigger for rebinning of spatial exposure inhomogeneities (CCD columns)
  logical             :: eekeep    ! trigger for checking rather than removing EXPOSU columns
!
! compile time parameters
!
  logical, parameter  :: tdefault = .FALSE. ! trigger for backward compatibility:
                                   ! if T then TIME has unrandomized times (as always in the past) and TIME_RND is randomized
                                   ! if F then TIME_RAW has unrandomized times and TIME is randomized
  logical, parameter  :: use_expo = .FALSE. ! trigger to use EXPOSU## extension in time randomization
! logical, parameter  :: ekeep = .FALSE. ! trigger for checking rather than removing EXPOSU columns
  logical, parameter  :: spr6672 = .TRUE. ! trigger for checking rather than removing EXPOSU columns
!-------------------------------------------------------------------------------
  write(errstr,*) 'EPEXPOSURE ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  single_ccd_mode = .TRUE.
  quadrantId = 0
  CCDId = 0
  ccdmode = 0
  expo_dlimap(1:64) = 0

!-------------------
! get all parameters
!-------------------

  npar = parameterCount("eventsets")
  write(errstr,*) 'Parameter eventsets has ',npar,' entries'
  call message(AppMsg,Noisy)
  if (npar .gt. 1) then
    call message(AppMsg,Sparse,'Can only handle one input file yet (may have several CCDs).')
  endif
  eventsets(0)=stringParameter("eventsets",0)
  write(errstr,*) 'Accessing ',trim(eventsets(0))
  call message(AppMsg,Noisy)

! trebin = integerParameter("rebinexposure")
  trebin = 1
  lrebin = (trebin > 1)
  if (lrebin) then
    call warning("rebinExposure","rebinning of EXPOSURE extension not yet enabled")
    lrebin = .FALSE.
  endif

  trandom = booleanParameter("randomizetime")
  if (trandom) then
!   call warning("randomizeTime","randomization of TIME not yet enabled")
!   trandom = .FALSE.
  endif

  espatial = booleanParameter("spatialexposure")
  if (espatial) then
    call warning("spatialExposure","determination of spatial (CCD column) exposure correction not yet enabled, dummy only.")
!   espatial = .FALSE.
  endif

  escreen = booleanParameter("screenexposure")
  eekeep = (.not.escreen)

  done = (.NOT.(espatial .OR. trandom .OR. escreen))

  if (.NOT. done) then

    infile = dataSet(trim(eventsets(0)),MODIFY,USE_ENVIRONMENT)
!   infile = dataSet(trim(eventsets(0)),MODIFY,LOW_MEMORY)
!   infile = dataSet(trim(eventsets(0)),MODIFY,HIGH_MEMORY)

!-------------------------------------------------
! define EXPOSURE related extension names
!-------------------------------------------------
    extname(0) =  "EXPOSURE"
    contime(0) =  "ONTIME"
    do CCDno = 1,nccd
      write(extname(CCDno),'(a,i2.2)') "EXPOSU" ,CCDno
      write(contime(CCDno),'(a,i2.2)') "ONTIME" ,CCDno
    enddo
    colname = (/'FRAMELIM','NDSCLIN1','NDSCLIN2','NDSCLIN3','NDSCLIN4'/)

    numberOfCCDs = 0
    do CCDno = 0,nccd
      write(errstr,*) 'checking existence of ', trim(extname(CCDno))
      call message(AppMsg,Verbose)
      hasExtension = hasBlock(infile,trim(extname(CCDno)))
      if (hasExtension) then
        hasExposure(CCDno) = .true.
	numberOfCCDs = numberOfCCDs+1
      endif
    enddo
    checkOFSTCOR = (numberOfCCDs.gt.1)

!-------------------
! delete extra columns
!-------------------
!!  if (escreen) then

    do CCDno = 0,nccd
      hasExtension = hasExposure(CCDno)
      if (hasExtension) then
          expo = table(infile,trim(extname(CCDno)))
          nrows = numberOfRows(expo)
          write(errstr,*) trim(extname(CCDno)),' has ',nrows,' number of rows'
          call message(AppMsg,Noisy)
!
! check presencw of OFSTCOR keywords
!   if  present: read OFSTCOR keywords
!                check whether they have same value
!   if not present: no check possible or required
!	
        if (hasAttribute(expo,"OFSTCOR1")) then
          ofstcor1(CCDno) = stringAttribute(expo,"OFSTCOR1")
        else
          ofstcor1(CCDno) = "None"
	endif
        if (hasAttribute(expo,"OFSTCOR4")) then
          ofstcor4(CCDno) = booleanAttribute(expo,"OFSTCOR4")
        else
          ofstcor4(CCDno) = .FALSE.
	endif
        if (escreen) then
          if (.not.eekeep) then
            do i = 0,4
              if (hascolumn(expo,trim(colname(i)))) then
                write(errstr,*) trim(colname(I)),' will be removed now'
                call message(AppMsg,Noisy)
!               expo_column = column(expo,trim(colname(i)))
                call deleteColumn(expo,trim(colname(i)))
              endif
            enddo
          else
            ndsclin1 = column(expo,"NDSCLIN1",READ)
            ndsclin2 = column(expo,"NDSCLIN2",READ)
            ndsclin3 = column(expo,"NDSCLIN3",READ)
            ndsclin4 = column(expo,"NDSCLIN4",READ)

            ndsclin1Col => int16Data(ndsclin1)
            ndsclin2Col => int16Data(ndsclin2)
            ndsclin3Col => int16Data(ndsclin3)
            ndsclin4Col => int16Data(ndsclin4)

            allocate(ndsclin(0:3,0:nrows-1))

            ndsclin(0,0:nrows-1) = ndsclin1Col(0:nrows-1)
            ndsclin(1,0:nrows-1) = ndsclin2Col(0:nrows-1)
            ndsclin(2,0:nrows-1) = ndsclin3Col(0:nrows-1)
            ndsclin(3,0:nrows-1) = ndsclin4Col(0:nrows-1)
      
            do irow = 0, nrows-1
              do i = 0, 3
                if (ndsclin(i,irow) /= 0) then
                  n = i * 16
                  do k = 1, 16
                    ii = k + n
                    ldisc = (ibits(ndsclin(i,irow),k-1,1) > 0)
                    if (ldisc) expo_dlimap(ii) = expo_dlimap(ii)+1
                  enddo
                endif
              enddo
            enddo

          endif
        endif ! escreen
        call release(expo)
      endif ! has exposure
    enddo

    mccd = 0
    warningOFSTCOR = .false.
    do CCDno = 0, nccd
      if (hasExposure(CCDno)) then
        mccd = mccd + 1
	if (mccd == 1) then
          ofstref1 = ofstcor1(CCDno)
          ofstref4 = ofstcor4(CCDno)
        else
          if (.not.warningOFSTCOR) warningOFSTCOR = .NOT.(trim(ofstcor1(CCDno))==trim(ofstref1))
	endif
	if (trim(ofstcor1(CCDno))=="with20aduImag") then
          if (.not.warningOFSTCOR) warningOFSTCOR = .NOT.(ofstcor4(CCDno).EQV.ofstref4)
          ofstref4 = ofstcor4(CCDno)
	endif
      endif
    enddo

    if (warningOFSTCOR) then
      call message(AppMsg,Verbose,' ')
      call warning("offsetCorrectionNotHomogeneous", &
              "offset corrections are different for the CCDs in event file.")
      call message(AppMsg,Verbose,'CCD          Correction Cal')
      do CCDno = 0, nccd
        if (hasExposure(CCDno)) then
          write(errstr,'(i3.2,1x,a20,1x,l1)') CCDno,trim(ofstcor1(CCDno)),ofstcor4(CCDno)
          call message(AppMsg,Verbose)
	endif
      enddo
      call message(AppMsg,Verbose,' ')
    endif

!!    endif  ! escreen



!-------------------------------------------------
! define DLIMAP related extension names
!-------------------------------------------------
    extname(0) =  "DLIMAP"  ! for intermediate event files
    isIntermediate = hasBlock(infile,trim(extname(0)))
!
    do CCDno = 1,nccd       ! for final merged event files
      write(extname(CCDno),'(a,i2.2)') "DLIMAP" ,CCDno
    enddo
    colname(0:1) = (/'DLIODF','DLISAS'/)
!
    lccd(0:nccd) = .false.   ! check whether extension is present in input file
    do CCDno = 0,nccd
      write(errstr,*) 'checking existence of ', trim(extname(CCDno))
      call message(AppMsg,Verbose)
      hasExtension = hasBlock(infile,trim(extname(CCDno)))
      if (hasExtension) then
        lccd(CCDno) = .true.
        dlimap = table(infile,trim(extname(CCDno)))
        nrows = numberOfRows(dlimap)
        write(errstr,*) trim(extname(CCDno)),' has ',nrows,' number of rows'
        call message(AppMsg,Noisy)
        quadrantId = (CCDno-1)/3
        do i = 0,1
          if (hascolumn(dlimap,trim(colname(i)))) then
            write(errstr,*) trim(colname(I)),' will be analysed now'
            call message(AppMsg,Noisy)
            dlicol(i) = column(dlimap,trim(colname(i)))
            outdliCol => int32Data(dlicol(i))
            do j = 0,EPN_MAX_Y-1
              dliarray(i+1,j+1,CCDNo) = outdlicol(j)
            enddo
          endif
        enddo
        if (hasAttribute(dlimap,"CCDMODE")) ccdmode = int16Attribute(dlimap,"CCDMODE")
        call release(dlimap)
      else
        dliarray(1:2,1:EPN_MAX_Y,CCDNo) = 0
      endif
    enddo
!
! now DLIARRAY contains all relevant MIP information collected by task "epframes":
!     DLIARRAY(1,X,CCD) = number of ODF rejected MIP columns for CCD
!     DLIARRAY(2,X,CCD) = number of SAS rejected MIP columns for CCD
!
    if (espatial) then
!
      ndsclinq_quad(0:3) = 0
      dliodf_tot_quad(0:3) = 0
      if (ccdmode .ge. 3) then
        single_CCD_mode = .true.    ! SW TI BU
        col_per_quad = EPN_MAX_X
        col_to_scale = EPN_MAX_X
      else                    
        single_CCD_mode = .false.   ! FF eFF LW
        col_per_quad = EPN_MAX_X*3
        col_to_scale = EPN_MAX_X
      endif
!
! DLI files show I*2 overflow if exposure is long
! proposed alogorithm needs to be recalibrated/redesigned
!
! append DLIEXPCO column of DLI-caused exposure correction factors
!
! works only for COLUMN-WISE MIP rejection (i.e. default)
!
! we need the following information:
!  ONTIME
!  FRAMETIM  -> these two give the number of frames
!  NDSCLINC  -> total number SAS per CCD
!  NDSCLINQ  -> total number onboard per quadrant
!  AN_DSLIN  -> with this number one can estimate the probability of
!               I*2 overflow in the DLIMAP, per 20 frames per quadrant
!
      call message(AppMsg,Verbose,' ')
      ccxdslin(0:nccd) = 0
      do CCDno = 0,nccd
        quadrantId = (CCDno-1)/3
        if (lccd(CCDno)) then
          if (CCDno .eq. 0) single_CCD_mode = .true. ! intermediate file
          write(errstr,*) 'updating ', trim(extname(CCDno))
          call message(AppMsg,Verbose)
          dlimap = table(infile,trim(extname(CCDno)))
!
          call message(AppMsg,Noisy,' ')
          if (hasAttribute(dlimap,"SUBMODE")) then
            submode = trim(stringAttribute(dlimap,"SUBMODE"))
            write(errstr,'(1x,a,a)') '  SubMode   =        ',trim(submode)
            call message(AppMsg,Noisy)
          else
            submode = ' '
            call message(AppMsg,Verbose,'SUBMODE attribute not found')
          endif
!
          if (hasAttribute(dlimap,"OBS_ID")) then
            ObsID   = trim(stringAttribute(dlimap,"OBS_ID"))
            write(errstr,'(1x,a,a)') '  ObsID     =        ',trim(ObsID)
            call message(AppMsg,Noisy)
          else
            ObsID = ' '
            call message(AppMsg,Verbose,'OBS_ID attribute not found')
          endif
!
! needs epframes-8.92.4 to read from DLIMAP (and not from PRIMARY header of "infile")
!
          if (hasAttribute(dlimap,"EXPIDSTR")) then
!         if (hasAttribute(infile,"EXPIDSTR")) then
            ExpIDstr = trim(stringAttribute(dlimap,"EXPIDSTR"))
!           ExpIDstr = trim(stringAttribute(infile,"EXPIDSTR"))
            write(errstr,'(1x,a,a)') '  ExpID     =        ',trim(ExpIDstr)
            call message(AppMsg,Noisy)
          else
            ExpIDstr = ' '
            call message(AppMsg,Verbose,'EXPIDSTR attribute not found')
          endif
!
          if (hasAttribute(dlimap,"REVOLUT")) then
            revolution = int16Attribute(dlimap,"REVOLUT")
            write(errstr,'(1x,a,i8)') '  Revolution =       ',revolution
            call message(AppMsg,Noisy)
          else
            revolution = 20 ! first EPIC-pn data
            call message(AppMsg,Verbose,'REVOLUT attribute not found')
          endif
!
          call message(AppMsg,Noisy,' ')
          if (hasAttribute(dlimap,"AN_SENDM")) then
            an_sendmode = trim(stringAttribute(dlimap,"AN_SENDM"))
            write(errstr,'(1x,a,a)') '  MIPcorrec =        ',trim(an_sendmode)
            call message(AppMsg,Noisy)
          else
            an_sendmode = 'MIP CORREC.2'  ! column-wise MIP rejection
            call message(AppMsg,Verbose,'AN_SENDM attribute not found')
          endif
!
          if (hasAttribute(dlimap,"AN_EAMIP")) then
            an_eamip = int16Attribute(dlimap,"AN_EAMIP")
            write(errstr,'(1x,a,i8)') '  EAMIP      =       ',an_eamip
            call message(AppMsg,Noisy)
          else
            an_eamip = 1 ! columns  rejected to either side of MIP
            call message(AppMsg,Verbose,'AN_EAMIP attribute not found')
          endif
!
          if (hasAttribute(dlimap,"AN_CHOP")) then
            an_chop = int16Attribute(dlimap,"AN_CHOP")
            write(errstr,'(1x,a,i8)') '  Chopper    =       ',an_chop
            call message(AppMsg,Noisy)
          else
            an_chop = 0 ! no chopper
            call message(AppMsg,Verbose,'AN_CHOP attribute not found')
          endif
!
          if (hasAttribute(dlimap,"AN_MAXFR")) then
            an_maxfrc = int16Attribute(dlimap,"AN_MAXFR")
            write(errstr,'(1x,a,i8)') '  Frames/counter =   ',an_maxfrc
            call message(AppMsg,Noisy)
          else
            an_maxfrc = 20 ! has it ever been changed ?
            call message(AppMsg,Verbose,'AN_MAXFR attribute not found')
          endif
!
          if (hasAttribute(dlimap,"FRAMETIM")) then
            frameTime = real32Attribute(dlimap,"FRAMETIM")
            write(errstr,'(A,f14.5,A)') '   FRAMETIM  = ',frameTime ,' [ms]'
            call message(AppMsg,Noisy)
          else
            frameTime = 73.36496 ! [ms] FF default
            call message(AppMsg,Verbose,'FRAMETIM attribute not found')
          endif
!
          if (hasAttribute(dlimap,"ONTIME")) then
            ontime = real32Attribute(dlimap,"ONTIME")
            write(errstr,'(A,f12.3,A)') '   ONTIME    = ', ONTIME,' [s]'
            call message(AppMsg,Noisy)
          else
            ontime = 1.0 ! [s]
            call message(AppMsg,Verbose,'ONTIME attribute not found')
          endif
!
!         ontime = ontime - fifo_loss
          if (hasAttribute(dlimap,"FIFOLOSS")) then
            fifo_loss = real32Attribute(dlimap,"FIFOLOSS")
            write(errstr,'(A,f12.3,A)') '   FIFOLOSS  = ',fifo_loss,' [s]'
            call message(AppMsg,Noisy)
          else
            fifo_loss = 0.0 ! [s]
            call message(AppMsg,Verbose,'FIFOLOSS attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"COUNTMOD")) then
            T_Counting_Mode = real32Attribute(dlimap,"COUNTMOD")
            write(errstr,'(A,f12.3,A)') '   COUNTMOD  = ',T_Counting_Mode,' [s]'
            call message(AppMsg,Noisy)
          else
            T_Counting_Mode = 0.0 ! s
            call message(AppMsg,Verbose,'COUNTMOD attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"NDSCLINA")) then ! NDSCLINQ rate
            ndsclina = real32Attribute(dlimap,"NDSCLINA")
            write(errstr,'(A,f12.3)') '   NDSCLINA  = ', NDSCLINA
            call message(AppMsg,Noisy)
          else
            ndsclina = 0 ! per exposure per quadrant, from PNAUX2
            call message(AppMsg,Verbose,'NDSCLINA attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"AN_DSLIN")) then   ! PNPMH1: F1330 F1375 F1420 F1465
            an_dslin = real32Attribute(dlimap,"AN_DSLIN")
            write(errstr,'(A,f12.3)') '   AN_DSLIN  = ', AN_DSLIN
            call message(AppMsg,Noisy)
          else
            an_dslin = 140.0 ! per AN_MAXFRC frames per quadrant, from PNPMH1
            call message(AppMsg,Verbose,'AN_DSLIN attribute not found')
          endif
!
! is this still true for chopper ? number of counter values * mean value of counter
!
          dslin_pmh(quadrantId) = ontime/frametime*1.d3/an_maxfrc*an_dslin*(1+AN_CHOP)
          write(errstr,'(A,f12.3)') '   DSLIN_PMH = ', dslin_pmh(quadrantId)
          call message(AppMsg,Noisy)
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"CCXDSLIN")) then   ! number during couting mode intervals
            ccxdslin(CCDno) = real32Attribute(dlimap,"CCXDSLIN")
!           write(errstr,'(A,i12)') '   DSLIN_CCX = ', CCXDSLIN(CCDno)
            write(errstr,'(A,2i8)') '   DSLIN_CCX = ', CCXDSLIN(CCDno),CCDno
            call message(AppMsg,Noisy)
          else
            ccxdslin(CCDno) = 0 ! total number during couting mode per quadrant, from PNCCX1
            call message(AppMsg,Verbose,'CCXDSLIN attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIMEAN2")) then
            dlimean2(ccdno) = real32Attribute(dlimap,"DLIMEAN2")
            write(errstr,'(A,f12.3,i4)') '   DLIMEAN   = ', DLIMEAN2(ccdno),CCDno
            call message(AppMsg,Noisy)
          else
            dlimean2(ccdno) = 0 ! mean value from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIMEAN2 attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIMEDI2")) then
            dlimedian2(ccdno) = real32Attribute(dlimap,"DLIMEDI2")
            write(errstr,'(A,f12.3,i4)') '   DLIMEDIAN = ', DLIMEDIAN2(ccdno),CCDno
            call message(AppMsg,Noisy)
          else
            dlimedian2(ccdno) = 0 ! median value from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIMEDI2 attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIMIN")) then
            dlimin(ccdno) = real32Attribute(dlimap,"DLIMIN")
            write(errstr,'(A,f12.3,i4)') '   DLIMIN    = ', DLIMIN(ccdno)
            call message(AppMsg,Noisy)
          else
            dlimin(ccdno) = 0 ! minimum value from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIMIN attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIMAX")) then
            dlimax(ccdno) = real32Attribute(dlimap,"DLIMAX")
            write(errstr,'(A,f12.3,i4)') '   DLIMAX    = ', DLIMAX(ccdno)
            call message(AppMsg,Noisy)
          else
            dlimax(ccdno) = 0 ! maximum value from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIMAX attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIADEV")) then
            dliadev(ccdno) = real32Attribute(dlimap,"DLIADEV")
            write(errstr,'(A,f12.3,i4)') '   DLIADEV   = ', DLIADEV(ccdno)
            call message(AppMsg,Noisy)
          else
            dliadev(ccdno) = 0 ! absolute deviation from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIADEV attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLISDEV")) then
            dlisdev(ccdno) = real32Attribute(dlimap,"DLISDEV")
            write(errstr,'(A,f12.3,i4)') '   DLISDEV   = ', DLISDEV(ccdno)
            call message(AppMsg,Noisy)
          else
            dlisdev(ccdno) = 0 ! standard deviation from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLISDEV attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLISKEW")) then
            dliskew(ccdno) = real32Attribute(dlimap,"DLISKEW")
            write(errstr,'(A,f12.3,i4)') '   DLISKEW   = ', DLISKEW(ccdno)
            call message(AppMsg,Noisy)
          else
            dliskew(ccdno) = 0 ! skew from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLISKEW attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"DLIKURT")) then
            dlicurt(ccdno) = real32Attribute(dlimap,"DLIKURT")
            write(errstr,'(A,f12.3,i4)') '   DLIKURT   = ', DLICURT(ccdno)
            call message(AppMsg,Noisy)
          else
            dlicurt(ccdno) = 0 ! skew from DLI file for inner CCD columns
            call message(AppMsg,Verbose,'DLIKURT attribute not found')
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"NDSCLINQ")) then
            ndsclinq = int32Attribute(dlimap,"NDSCLINQ")
            write(errstr,'(A,I8)') '   NDSCLINQ  = ', NDSCLINQ
            call message(AppMsg,Noisy)
          else
            ndsclinq = 0 ! per exposure per quadrant, from PNAUX2
            call message(AppMsg,Verbose,'NDSCLINQ attribute not found')
          endif
          call message(AppMsg,Noisy,' ')
          dlimap_overflow = (ndsclinq/(1.0*col_per_quad) .gt. 64000)
          if (dlimap_overflow) then
            call warning("spatialDLIgeneral", &
              "discarded line files may be affected by I*2 overflow due to high rate or long exposure.")
          endif
!-------------------------------------------------------------------------------
          if (hasAttribute(dlimap,"NDSCLINC")) then
            ndsclinc = int32Attribute(dlimap,"NDSCLINC")
            write(errstr,'(A,I8)') '   NDSCLINC  = ', NDSCLINC
            call message(AppMsg,Noisy)
          else
            ndsclinc = 0 ! per exposure per quadrant, from SAS on-fround
            call message(AppMsg,Verbose,'NDSCLINC attribute not found')
          endif
!-------------------------------------------------------------------------------
          nframe = NINT(ontime/frameTime*1.d3) ! number of read-out frames
!
          if (hasColumn(dlimap,"DLIEXPCO")) then
            dliexpco = column(dlimap,"DLIEXPCO",MODIFY)
          else
            dliexpco = addColumn(dlimap,"DLIEXPCO",Real32,"")
          endif
          dliexpcoCol => real32Data(dliexpco)

!         if (single_CCD_mode) then
            dliexpcoCol(0:EPN_MAX_Y-1) = 0.0
            dli_total = 0
            dliodf_total = 0
            dlisas_total = 0
            do j = 0,EPN_MAX_X-1  ! [1..64]
              dliodf_total = dliodf_total + dliarray(1,j+1,CCDno) ! from DLI for this CCD
              dlisas_total = dlisas_total + dliarray(2,j+1,CCDno) ! from SAS for this CCD
            enddo
            if (ccdno .gt. 0) then
              dliodf_tot_quad(quadrantId) = dliodf_tot_quad(quadrantId)+dliodf_total
              ndsclinq_quad(quadrantId) = ndsclinq
            else
              dliodf_tot_quad(quadrantId) = dliodf_tot_quad(quadrantId)+dliodf_total
              ndsclinq_quad(quadrantId) = ndsclinq
            endif
            do j = 1,EPN_MAX_X
              write(errstr,'(A,i4,3I8,f10.6)') '   NDSCLIN  = ', &
                    j,dliarray(1,j,CCDno),dliarray(2,j,CCDno),expo_dlimap(j), &
                    expo_dlimap(j)/(1.e-7+dliarray(2,j,CCDno))
              call message(AppMsg,Verbose)
            enddo
            call message(AppMsg,Verbose,' ')
!
! issue a warning if DLIODF columns per quadrant are significantly
! lower than the NDSCLINQ columns (not affected by I*2 overflow!)
!
!           call message(AppMsg,Noisy,' ')
            dli_total = dliodf_total+dlisas_total
!           average = (1.0*dli_total)/col_per_quad
            average = (1.0*dli_total)/col_to_scale
            write(errstr,'(A,f12.3)') '   AVERAGE   = ',average
            call message(AppMsg,Noisy)
            write(errstr,'(A,f12.3)') '   NFRAMES   = ',nframe
            call message(AppMsg,Noisy)
            write(errstr,'(A,f14.5)') '   FRACEXP   = ',1.d0-(average/3/nframe)*20
            call message(AppMsg,Noisy)
            call message(AppMsg,Noisy,' ')
            do j = 0,EPN_MAX_X-1
              dliexpcoCol(j) = 1.d0-(dliarray(2,j+1,CCDno)-average)/average
            enddo
!         else
!           dliexpcoCol(0:EPN_MAX_Y-1) = 1.0
!         endif
          call release(dliexpco)
          call release(dlimap)
        endif
      enddo
!
! if SKEW < 0: tail to left side of distribution
! if KURTOSIS < 0: broad distribution (platycurtic) ; > 0: peaked distribution (leptokurtic)
!
!     call message(AppMsg,Sparse,' ')
      call message(AppMsg,Sparse,'Summary of discarded line information (MIP rejection):')

      lfirst = .true.
      do CCDNo = 0,nccd
        if (dlimax(ccdno).gt.0) then
          if (lfirst) then
            call message(AppMsg,Sparse,' ')
            call message(AppMsg,Sparse,' CCD  Minimum Maximum     Mean     Median       Adev       Sdev       Skew    Kurtosis')
            lfirst = .false.
          endif
          write(errstr,'(i4,2i8,2f11.2,2f11.3,2f11.4)') ccdno, &
             NINT(dlimin(ccdno)),NINT(dlimax(ccdno)),dlimean2(ccdno),dlimedian2(ccdno),&
             dliadev(ccdno),dlisdev(ccdno),dliskew(ccdno),dlicurt(ccdno)
          call message(AppMsg,Sparse)
        endif
      enddo
      call message(AppMsg,Sparse,' ')

      nval = 0
      avemean2 = 0.0
      medmean3 = 0.0
      rmin = 0.0
      rmax = 0.0
      adev = 0.0
      sdev = 0.0
      var  = 0.0
      skew = 0.0
      curt = 0.0
      do CCDno = 1,nccd
        if (dlimean2(CCDno).gt.0) then
          nval = nval+1
          xval(CCDno) = dlimean2(CCDno)
        endif
      enddo
      if (nval .gt. 0) then
        rmin = MINVAL(xval(1:nval))
        rmax = MAXVAL(xval(1:nval))
        call NR_moment(xval,nval,avemean2,adev,sdev,var,skew,curt,ierr)
        call ep_median0(xval,nval,medmean3)
        write(errstr,'(1x,a,i3,4f12.3)') '  N, Mean_median of means, Min,Max   = ', &
                    nval,avemean2,medmean3,rmin,rmax
        call message(AppMsg,Sparse)
        write(errstr,'(1x,a,2f12.3,2f12.4)') &
               '   adev,sdev,skew,kurtosis           =    ',adev,sdev,skew,curt
        call message(AppMsg,Sparse)
      endif

      nval = 0
      avemedian2 = 0.0
      medmedian3 = 0.0
      rmin = 0.0
      rmax = 0.0
      adev = 0.0
      sdev = 0.0
      var  = 0.0
      skew = 0.0
      curt = 0.0
      do CCDno = 1,nccd
        if (dlimedian2(CCDno).gt.0) then
          nval = nval+1
          xval(CCDno) = dlimedian2(CCDno)
        endif
      enddo
      if (nval .gt. 0) then
        rmin = MINVAL(xval(1:nval))
        rmax = MAXVAL(xval(1:nval))
        call NR_moment(xval,nval,avemedian2,adev,sdev,var,skew,curt,ierr)
        call ep_median0(xval,nval,medmedian3)
        write(errstr,'(1x,a,i3,4f12.3)') '  N, Mean_median of medians, Min,Max = ', &
                    nval,avemedian2,medmedian3,rmin,rmax
        call message(AppMsg,Sparse)
        write(errstr,'(1x,a,2f12.3,2f12.4)') &
               '   adev,sdev,skew,kurtosis           =    ',adev,sdev,skew,curt
        call message(AppMsg,Sparse)
      endif

!     call message(AppMsg,Sparse,' ')
      do quadrantId = 0,3
        ic0 = quadrantId*3+1 ! CCD nr of CCD0 in quadrant
        ic1 = quadrantId*3+2 ! CCD nr of CCD1 in quadrant
        ic2 = quadrantId*3+3 ! CCD nr of CCD2 in quadrant
        write(errstr,'(A,3(I4,L2))') ' CCDs: ',IC0,LCCD(IC0),IC1,LCCD(IC1),IC2,LCCD(IC2)
        call message(AppMsg,Noisy)
!
        if (lccd(ic0).or.lccd(ic1).or.lccd(ic2)) then ! if any of CCDs present:
!
          call message(AppMsg,Sparse,' ')
          diff = dliodf_tot_quad(quadrantId)-ndsclinq_quad(quadrantId) ! difference DLI-PNAUX2
          write(errstr,*) 'Quadrant ',quadrantID,':  PMH:', &
            NINT(dslin_pmh(quadrantId)),'  PNAUX:', &
            ndsclinq_quad(quadrantId),    '  DLI:', &
            dliodf_tot_quad(quadrantId),  '  CCX:', &
            ccxdslin(ic0), '  DIFF:',diff-ccxdslin(ic0)
          call message(AppMsg,Sparse)
!         call message(AppMsg,Verbose,' ')
!
          dlimap_overflow = (diff-ccxdslin(ic0) .lt. -15000)
          if (dlimap_overflow) then
            noverflow(quadrantId) = NINT(ABS(1.0*diff)/65536)
            write(errstr,*) 'discarded line files may be affected by I*2 overflow in quadrant',quadrantID
            call warning('spatialDLIquadrant', trim(errstr))
            if (noverflow(quadrantId) .eq. 1) then
              write(errstr,'(1x,a,i1,a,i5,a)') &
                 'The PNAUX-DLI difference in Q',quadrantId,' corresponds to about',&
                  noverflow(quadrantId),' I*2 overflow'
            else
              write(errstr,'(1x,a,i1,a,i5,a)') &
                 'The PNAUX-DLI difference in Q',quadrantId,' corresponds to about',&
                  noverflow(quadrantId),' I*2 overflows'
            endif
            call message(AppMsg,Sparse)
!           call message(AppMsg,Sparse,' ')
          endif
!
          dlimap_overflow = (diff-ccxdslin(ic0) .gt. +45000)
          if (dlimap_overflow) then
            moverflow(quadrantId) = NINT(ABS(1.0*diff)/65536)
            write(errstr,*) 'discarded line files show significant deficiency (low quality) in quadrant',quadrantID
            call warning('spatialDLIquality', trim(errstr))
            if (moverflow(quadrantId) .eq. 1) then
              write(errstr,'(1x,a,i1,a,i5,a)') &
                 'The DLI-PNAUX difference in Q',quadrantId,' corresponds to about',&
                  moverflow(quadrantId),' I*2 overflow'
            else
              write(errstr,'(1x,a,i1,a,i5,a)') &
                 'The DLI-PNAUX difference in Q',quadrantId,' corresponds to about',&
                  moverflow(quadrantId),' I*2 overflows'
            endif
            call message(AppMsg,Sparse)
            call message(AppMsg,Sparse,' ')
          endif
        endif ! at least 1 ccd of quadrant present
  
        if (noverflow(quadrantId).gt.0) then
          continue  ! this is the place for DLI restoration
        endif
      enddo

!      do ccdNo = 0,nccd
!        write(errstr,*) ccdNo,ccxdslin(ccdNo)
!        call message(AppMsg,Sparse)
!      enddo

    endif

!-----------------------------------------------------------------------
! do randomization in TIME within one readout frame
! NOTE: it is assumed that the files are fractionally ordered,
!   i.e. for each (sorted) CCDNR value the events are sorted in TIME;
!   this is the case for normal EPIC-pn SAS event files (PPS, epchain)
!-----------------------------------------------------------------------

    extname(0) =  "EXPOSURE"
    do CCDno = 1,nccd
      write(extname(CCDno),'(a,i2.2)') "EXPOSU" ,CCDno
    enddo
    colname = (/'FRAMELIM','NDSCLIN1','NDSCLIN2','NDSCLIN3','NDSCLIN4'/)

    do CCDno = 0,nccd
      hasExtension = hasBlock(infile,trim(extname(CCDno)))
      if (hasExtension) then
        expo = table(infile,trim(extname(CCDno)))
        call setAttribute(expo, "RAND_TIM", 0, "", "1: randomized time 0: not randomized")
        call release(expo)
      endif
    enddo

!!  if (trandom) then
      ext_events = 'EVENTS'
      write(errstr,*) 'checking existence of ', trim(ext_events)
      call message(AppMsg,Verbose)
      hasExtension = hasBlock(infile,trim(ext_events))
      if (hasExtension) then
        events = table(infile,trim(ext_events))
	call setAttribute(events,"WarnOfst",warningOFSTCOR,comment="Inhomogeneous offset corrections")
      if (trandom) then
        if (hasAttribute(infile,"SUBMODE")) then
          submode = trim(stringAttribute(infile,"SUBMODE"))
          write(errstr,'(1x,a,t39,a)') 'SubMode:     ',trim(submode)
          call message(AppMsg,Verbose)
!
! no recipe yet for FAST modes (TI+BU), so issue warning and do nothing here:
!
          if (trim(submode) .eq. trim(modeId(FAST_TIMING)) .or. &
              trim(submode) .eq. trim(modeId(FAST_BURST))  ) then
            fastmode = .TRUE.
            call warning("randomizeTime", &
              "randomization of TIME not yet enabled for FAST modes (TI,BU).")
          else
            fastmode = .FALSE.
          endif
        else
          submode = ' '
          call message(AppMsg,Verbose,'SUBMODE attribute not found')
        endif
!
        if (hascolumn(events,'TIME').and.hascolumn(events,'CCDNR').and.(.NOT.fastmode)) then
          nrows = numberOfRows(events)
          if (nrows .eq. 0) then
            emptyEvents = .true.  ! e.g., 0572_0134522201_PNU003
          else
            emptyEvents = .false.          
          endif

          if (tdefault) then
            if (hascolumn(events,'TIME_RND')) then
!             nothing to be done, already backup column present
            else
              call copyColumn(events,column(events,'TIME'),'TIME_RND')
            endif
          else
            if (hascolumn(events,'TIME_RAW')) then
!             nothing to be done, already backup column present
            else
              call copyColumn(events,column(events,'TIME'),'TIME_RAW')
            endif
          endif
!
! always use the original TIME column for randomization as it already
! contains the proper SAS fine time corrections
!
          if (tdefault) then
            time_raw  => real64Data(column(events,'TIME'))
            time_rnd  => real64Data(column(events,'TIME_RND'))
          else
            time_raw  => real64Data(column(events,'TIME_RAW'))
            time_rnd  => real64Data(column(events,'TIME'))
          endif
          ccdnr => int8Data(column(events,'CCDNR'))
!
         if (.not.emptyEvents) then
!
! read start time (first event), however, not used later at the moment
!
          t_raw = time_raw(0)
          t_rnd = time_rnd(0)
!
! check which blocks of events belong to a certain CCD
!
          lccd(1:nccd) = .false.
          ic_b(1:nccd) = 0
          ic_e(1:nccd) = 0
          ccxdslin(0:nccd) = 0
!
          ic = ccdnr(0)
          ic_b(ic) = 1
          lccd(ic) = .true.
          do irow = 1, nrows-1
            if (ccdnr(irow).gt.ic) then
              ic_e(ic) = irow-2
              ic = ccdnr(irow)
              ic_b(ic) = irow-1
              lccd(ic) = .true.
            endif
          enddo
          ic_e(ic) = nrows
          do ic = 1,nccd
            write(errstr,'(a,i2,a,i8,a,i8)') 'CCD ',ic,': from event ',&
                ic_b(ic),' to ',ic_e(ic)
            call message(AppMsg,Noisy)
          enddo
!         nullify(ccdnr)

          do CCDno = 0,nccd
            if (CCDno.ge.1) then
              if (.NOT.lccd(ccdno)) cycle ! no events with this CCD number
            endif
            write(errstr,*) 'checking existence of ', trim(extname(CCDno))
            call message(AppMsg,Verbose)
            hasExtension = hasBlock(infile,trim(extname(CCDno)))
            if (hasExtension) then
              expo = table(infile,trim(extname(CCDno)))
              mrows = numberOfRows(expo)
              write(errstr,*) trim(extname(CCDno)),' has ',mrows,' number of rows'
              call message(AppMsg,Noisy)
!
! TIMEDEL is *not* the proper quantity for randomization at it gives only
! the exposure time under the PSF but does not include out-of-time events,
! it has to be replaced by the total frame time minus real dead time (later).
! But unfortunately/apparently people WANT to deal with the "PSF time",
! so they get what they want ...
!
! TIME is the central time of the readout interval, so randomize within
! this interval, +-0.5*TIMEDEL
!
              if (hasAttribute(expo,"TIMEDEL")) then
                timedelta = real32Attribute(expo,"TIMEDEL")
                write(errstr,'(1x,a,t39,f10.6,a)') 'TimeDel:',timedelta,' [s]'
                call message(AppMsg,Verbose)
               else
                submode = ' '
                call message(AppMsg,Verbose,'TIMEDEL attribute not found')
                timedelta = 0.0d0
              endif
              if (use_expo) then
                time_exp => real64Data(column(expo,'TIME'))
                t_exp = time_exp(0)
              endif
!
              call CAL_setState(instrument=EPN, randomize = .true.)
              if (CCDno.eq.0) then
                if (spr6672) then
                  allocate(dummy(0:nrows-1))
                  dummy(0:nrows-1) = 0
                  call CAL_setRandom(-0.5d0,+0.5d0) ! SAS_RAND_SEED
                  call CAL_integerToReal(dummy,randomnumber)
                  do irow = 0,nrows-1
                    time_rnd(irow) = time_raw(irow)+randomnumber(irow)*timedelta
                  enddo
                  deallocate(dummy)
                  call CAL_releaseMemory(randomnumber)
                else
                  do irow = 0,nrows-1
                    call random_number(rr)
                    time_rnd(irow) = time_raw(irow)+(rr-0.5d0)*timedelta
                  enddo
                endif
              else
                write(errstr,'(1x,a,t39,2i9)') 'CCD index range:',ic_b(CCDno),ic_e(CCDno)
                call message(AppMsg,Noisy)
                if (spr6672) then
                  mrows = ic_e(CCDno)-ic_b(CCDno)+1
                  allocate(dummy(0:mrows-1))
                  dummy(0:mrows-1) = 0
                  call CAL_setRandom(-0.5d0,+0.5d0) ! SAS_RAND_SEED
                  call CAL_integerToReal(dummy,randomnumber)
                  do irow = ic_b(CCDno)-1,ic_e(CCDno)-1
                    ir = irow-ic_b(CCDno)+1
                    time_rnd(irow) = time_raw(irow)+randomnumber(ir)*timedelta
                  enddo
                  deallocate(dummy)
                  call CAL_releaseMemory(randomnumber)
                else
                  do irow = ic_b(CCDno)-1,ic_e(CCDno)-1
                    call random_number(rr)
                    time_rnd(irow) = time_raw(irow)+(rr-0.5d0)*timedelta
                   enddo
                endif
              endif
              call CAL_setState(instrument=EPN, randomize = .false.)
              call setAttribute(expo, "RAND_TIM", 1, "", "1: randomized time 0: not randomized")
              call release(expo)
            endif
          enddo
!         nullify(time_exp)
!         nullify(time_rnd)
!         nullify(time_raw)
         else  ! empty EVENTS extension
	   call warning("emptyEvents","empty EVENTS extension, no time values for randomization")
         endif ! empty EVENTS extension
        else ! does not have TIME and/or CCDNR columns

        endif

    endif  !! trandom

        call release(events)

      else ! does not have EVENTS

      endif

!!  endif  !! trandom

!-------------------------------------------------
! write out modified event file
!-------------------------------------------------
    call message(AppMsg,Verbose,'Writing output file')
    call release(infile)

  else
    call message(AppMsg,Sparse,'Nothing to be done for input file(s).')
  endif


  end subroutine epexposure

end module epexposure_mod
